"""
Logging module
"""

from .logger_config_base import (
    LoggerBase,
    LoggerConfig,
    LogLevel,
    Writable,
    SinkInput,
    SinkAccepted,
    RotationType,
    RetentionType,
    CompressionType,
    FilterType,
    OutputSink,
)

__all__ = [
    "LoggerBase",
    "LoggerConfig",
    "LogLevel",
    "Writable",
    "SinkInput",
    "SinkAccepted",
    "RotationType",
    "RetentionType",
    "CompressionType",
    "FilterType",
    "OutputSink",
]



"""
Exceptions submodule
"""

from .http_exceptions import (
    BaseHttpException,
    StaticAssetNotFound,
    AborterException,
    abort,
    html_abort,
)

from .runtime_exceptions import (
    CustomException,
    MethodNotControllerMethod,
    UnexpectedDecorator,
)

from .exception_handler import ExceptionHandler, handles
from werkzeug.exceptions import NotFound, MethodNotAllowed

__all__ = [
    "CustomException",
    "BaseHttpException",
    "StaticAssetNotFound",
    "AborterException",
    "abort",
    "html_abort",
    "MethodNotControllerMethod",
    "UnexpectedDecorator",
    "ExceptionHandler",
    "handles",
    "NotFound",
    "MethodNotAllowed",
]



"""
Common and supported media types for incoming requests and responses
"""

from enum import StrEnum


class MediaType(StrEnum):
    APPLICATION_X_WWW_FORM_URLENCODED = "application/x-www-form-urlencoded"
    MULTIPART_FORM_DATA = "multipart/form-data"
    APPLICATION_JSON = "application/json"
    APPLICATION_PROBLEM_JSON = "application/problem+json"
    APPLICATION_XML = "application/xml"
    TEXT_XML = "text/xml"
    TEXT_PLAIN = "text/plain"
    TEXT_HTML = "text/html"
    APPLICATION_OCTET_STREAM = "application/octet-stream"
    IMAGE_PNG = "image/png"
    IMAGE_JPEG = "image/jpeg"
    IMAGE_GIF = "image/gif"
    APPLICATION_PDF = "application/pdf"
    APPLICATION_X_NDJSON = "application/x-ndjson"
    APPLICATION_CSV = "application/csv"
    TEXT_CSV = "text/csv"
    APPLICATION_YAML = "application/yaml"
    TEXT_YAML = "text/yaml"
    APPLICATION_GRAPHQL = "application/graphql"
    NO_CONTENT = "empty"



"""
CLI controller module for Patera.
"""

import inspect
import asyncio
from typing import TYPE_CHECKING, Any, Callable, cast
from functools import wraps

from ..utilities import run_sync_or_async

if TYPE_CHECKING:
    from ..patera import Patera


class CLIController:
    """
    Base class for CLI controllers.
    This class automatically registers methods decorated with @command and @argument
    as CLI commands in the provided Patera app instance.
    """

    def __init__(self, app: "Patera"):
        self._app: "Patera" = app
        self._cli_commands: dict[str, Callable] = {}
        self._cli_commands_help: dict[str, str] = {}
        self._ctrl_name: str = self.__class__.__name__
        self._register_commands()

    def _register_commands(self):
        for attr_name in dir(self):
            method = getattr(self, attr_name)
            if callable(method):
                command = getattr(method, "cli_command", {})
                if command.get("is_cli_command", False):
                    self._register_command(method, command)

    def _register_command(self, method: Callable, command: dict):
        self._cli_commands[method.__name__] = method
        self._cli_commands[cast(str, command.get("command_name"))] = method
        self._cli_commands_help[method.__name__] = command.get("help", "")

    def find_method(self, method: str) -> Callable | None:
        return self._cli_commands.get(method, None)

    def run_command(self, method: Callable, *args, **kwargs):
        try:
            converted_values = self.prepare_cli_args(
                method, kwargs.get("command_args", [])
            )
            return asyncio.run(run_sync_or_async(method, *converted_values))
        except Exception as exc:
            print(f"Failed to run method: {method.__name__}")
            print(
                "Help: ",
                self._cli_commands_help.get(method.__name__, "No help string provided"),
            )
            print(exc)

    def convert_value(self, value: str, annotation: type) -> Any:
        """
        Convert a CLI string value to the annotated type.
        """
        if annotation is inspect.Parameter.empty or annotation is str:
            return value

        if annotation is int:
            return int(value)

        if annotation is float:
            return float(value)

        if annotation is bool:
            lowered = value.strip().lower()
            if lowered in {"true", "1", "yes", "y", "on"}:
                return True
            if lowered in {"false", "0", "no", "n", "off"}:
                return False
            raise ValueError(f"Cannot convert '{value}' to bool")

        return annotation(value)

    def prepare_cli_args(self, method, raw_args: list[str]) -> list[Any]:
        """
        Convert raw CLI args to properly typed args based on method annotations.
        """
        sig = inspect.signature(method)
        params = list(sig.parameters.values())
        converted = []
        raw_index = 0

        for param in params:
            if param.name == "self":
                continue

            if raw_index >= len(raw_args):
                if param.default is not inspect.Parameter.empty:
                    converted.append(param.default)
                    continue
                raise ValueError(f"Missing required argument: {param.name}")

            raw_value = raw_args[raw_index]
            converted.append(self.convert_value(raw_value, param.annotation))
            raw_index += 1

        if raw_index < len(raw_args):
            raise ValueError(f"Too many arguments provided: {raw_args[raw_index:]}")

        return converted

    @property
    def app(self) -> "Patera":
        """Returns the Patera app instance."""
        return self._app

    @property
    def ctrl_name(self) -> str:
        return self._ctrl_name

    def set_ctrl_name(self, name: str) -> None:
        self._ctrl_name = name


def command(command_name: str, help: str = ""):

    def decorator(func):
        @wraps(func)
        async def wrapper(self: "CLIController", *args, **kwargs):
            return await run_sync_or_async(func, self, *args, **kwargs)

        attr = getattr(wrapper, "cli_command", {})
        attr["is_cli_command"] = True
        attr["command_name"] = command_name
        attr["help"] = help
        setattr(wrapper, "cli_command", attr)
        return wrapper

    return decorator



"""
Response class. Holds all information regarding responses to individual requests
"""

from collections.abc import AsyncIterable, Iterable
from typing import (
    Any,
    Optional,
    TYPE_CHECKING,
    Self,
    TypeVar,
    Generic,
    Type,
    cast,
    AsyncIterator,
)

from .media_types import MediaType
from .utilities import run_sync_or_async
from .http_statuses import HttpStatus

if TYPE_CHECKING:
    from .patera import Patera
    from .request import Request

U = TypeVar("U")


class Response(Generic[U]):
    """
    Response class of application. Holds all data (headers, body, status_code of the response)
    Example return of route handler:
    ```
    return res.json({"message": "My message", "status": "some status"}).status(200)
    ```
    """

    def __init__(self, app: "Patera", request: "Request") -> None:
        self._app = app
        self._request = request
        self.status_code: int | HttpStatus = HttpStatus.OK  # default status code is 200
        self.headers: dict = {}
        self.body: Optional[U] = None
        self.render_engine = self._app.jinja_environment
        self._zero_copy = None
        self._expected_body_type: Optional[Type[Any]] = None

        self._stream: Optional[AsyncIterable[bytes] | Iterable[bytes]] = None

    def status(self, status_code: int | HttpStatus) -> Self:
        """
        Sets status code of response
        """
        self.status_code = status_code
        return self

    def redirect(
        self, location: str, status_code: int | HttpStatus = HttpStatus.SEE_OTHER
    ) -> Self:
        """
        Redirects the client to the provided location
        """
        self.set_header("location", location)
        self.status_code = status_code
        self.body = None
        return self

    def json(self, data: Any) -> Self:
        """
        Sets data to response body and creates appropriate
        response headers. Sets default response status to 200
        """
        self.headers["content-type"] = MediaType.APPLICATION_JSON.value
        self.body = data
        return self

    def no_content(self) -> Self:
        """
        Returns a response with no content (body) and status 204
        """
        self.body = None
        self.status(204)
        return self

    def text(self, text: str) -> Self:
        """
        Creates text response with text/html content-type
        """
        self.body = cast(U, text.encode("utf-8"))
        self.status(HttpStatus.OK)
        return self

    async def html_from_string(
        self, text: str, context: Optional[dict[str, Any]] = None
    ) -> Self:
        """
        Creates text response with text/html content-type
        """
        if context is None:
            context = {}
        for method in self.app.global_context_methods:
            additional_context = await run_sync_or_async(method)
            if not isinstance(additional_context, dict):
                raise ValueError(
                    "Return of global context method must be off type dictionary"
                )
            context.update(additional_context)
        self.headers["content-type"] = "text/html"
        context["app"] = self.app
        context["url_for"] = self.app.url_for
        context["request"] = self._request
        context["attribute"] = getattr
        context["len"] = len
        rendered = await self.render_engine.from_string(text).render_async(
            **context
        )  # self.render_engine.from_string(text).render(**context)
        # self.body = text.encode("utf-8")
        self.body = cast(U, rendered.encode("utf-8"))
        self.status(HttpStatus.OK)
        return self

    async def html(
        self, template_path: str, context: Optional[dict[str, Any]] = None
    ) -> Self:
        """
        Renders html template and creates response with text/html content-type
        and default status code 200

        template_path: relative path of template inside the templates folder
        context: dictionary with data used in the template
        """

        if context is None:
            context = {}

        for method in self.app.global_context_methods:
            additional_context = await run_sync_or_async(method)
            if not isinstance(additional_context, dict):
                raise ValueError(
                    "Return of global context method must be off type dictionary"
                )
            context = {**context, **additional_context}
        context["app"] = self.app
        context["url_for"] = self.app.url_for
        context["request"] = self._request
        context["attribute"] = getattr
        context["len"] = len

        template = self.render_engine.get_template(template_path)
        rendered = await template.render_async(**context)
        self.headers["content-type"] = "text/html"
        self.body = cast(U, rendered.encode("utf-8"))
        self.status(HttpStatus.OK)
        return self

    def send_file(self, body, headers) -> Self:
        """
        For sending files
        Sets correct headers and body of the response
        """
        for k, v in headers.items():
            self.headers[k] = v
        self.body = body
        return self

    def set_header(self, key: str, value: str) -> Self:
        """
        Sets or updates a header in the response.

        key: Header name
        value: Header value
        """
        self.headers[key.lower()] = value
        return self

    def set_headers(self, headers: dict[str, str]) -> Self:
        """
        Sets multiple headers in the response.

        headers: Dictionary of headers to set
        """
        for key, value in headers.items():
            self.set_header(key, value)
        return self

    def set_cookie(
        self,
        cookie_name: str,
        value: str,
        max_age: int | None = None,
        path: str = "/",
        domain: str | None = None,
        secure: bool = False,
        http_only: bool = True,
    ) -> Self:
        """
        Sets a cookie in the response.

        cookie_name: Cookie name
        value: Cookie value
        max_age: Max age of the cookie in seconds (optional)
        path: Path where the cookie is available (default "/")
        domain: Domain where the cookie is available (optional)
        secure: If True, the cookie is only sent over HTTPS (default False)
        http_only: If True, the cookie is inaccessible to JavaScript (default True) <-- MORE SECURE
        """
        cookie_parts = [f"{cookie_name}={value}"]

        if max_age is not None:
            cookie_parts.append(f"Max-Age={max_age}")
        if path:
            cookie_parts.append(f"Path={path}")
        if domain:
            cookie_parts.append(f"Domain={domain}")
        if secure:
            cookie_parts.append("Secure")
        if http_only:
            cookie_parts.append("HttpOnly")

        cookie_header = "; ".join(cookie_parts)
        if "set-cookie" in self.headers:
            self.headers["set-cookie"] += f", {cookie_header}"
        else:
            self.headers["set-cookie"] = cookie_header

        return self

    @property
    def content_type(self) -> str:
        """
        Gets the Content-Type header of the response.
        content_type: The MIME type to set as the Content-Type
        """
        return cast(str, self.headers.get("content-type", None))

    def delete_cookie(
        self, cookie_name: str, path: str = "/", domain: Optional[str] = None
    ) -> Self:
        """
        Deletes a cookie by setting its Max-Age to 0.

        cookie_name: Cookie name
        path: Path where the cookie was available (default "/")
        domain: Domain where the cookie was available (optional)
        """
        cookie_parts = [f"{cookie_name}=", "Max-Age=0", f"Path={path}"]

        if domain:
            cookie_parts.append(f"Domain={domain}")

        cookie_header = "; ".join(cookie_parts)
        if "set-cookie" in self.headers:
            self.headers["set-cookie"] += f", {cookie_header}"
        else:
            self.headers["set-cookie"] = cookie_header

        return self

    def set_zero_copy(self, data) -> Self:
        """Sets zero copy data for range responses"""
        self._zero_copy = data
        return self

    def _set_expected_body_type(self, t: Optional[Type[Any]]) -> None:
        self._expected_body_type = t

    def expected_body_type(self) -> Optional[Type[Any]]:
        return self._expected_body_type

    async def send(self, message: dict) -> None:
        return await self._request.send(message)

    def stream(self, iterable: AsyncIterable[bytes] | Iterable[bytes]) -> Self:
        """
        Method for streaming response

        iterable: async or sync iterable yielding bytes (or bytearray/str)
        content_type: MIME type of the stream
        """
        self._stream = iterable
        self.headers["content-type"] = MediaType.APPLICATION_OCTET_STREAM.value
        # Body must be None so the app knows to stream
        self.body = None
        return self

    def stream_text(self, iterable: AsyncIterable[str] | Iterable[str]) -> Self:
        """
        Method for streaming text response
        """

        async def _aiter() -> AsyncIterator[bytes]:
            if hasattr(iterable, "__aiter__"):
                async for chunk in iterable:  # type: ignore[attr-defined]
                    yield str(chunk).encode("utf-8")
            else:
                for chunk in iterable:  # type: ignore[operator]
                    yield str(chunk).encode("utf-8")

        self._stream = _aiter()
        self.headers["content-type"] = MediaType.TEXT_PLAIN.value + "; charset=utf-8"
        self.body = None
        return self

    @property
    def stream_iterable(self) -> Optional[AsyncIterable[bytes] | Iterable[bytes]]:
        return self._stream

    @property
    def is_streaming(self) -> bool:
        return self._stream is not None

    @property
    def zero_copy(self):
        """Returns zero copy data"""
        return self._zero_copy

    @property
    def app(self):
        """
        Returns application reference
        """
        return self._app



# request.py
# pylint: disable=C0116

from __future__ import annotations

import re
import json
import base64
from io import BytesIO
from urllib.parse import parse_qs
from typing import Callable, Any, Union, TYPE_CHECKING, Mapping, cast

import python_multipart as pm
from pydantic_core import core_schema

from .response import Response

if TYPE_CHECKING:
    from .patera import Patera


def extract_boundary(content_type: str) -> str:
    match = re.search(r'boundary="?([^";]+)"?', content_type)
    if not match:
        raise ValueError("No boundary found in Content-Type")
    return match.group(1)


class UploadedFile:
    """
    Wrapper around an in-memory/temporary file.
    """

    def __init__(self, filename: str, content: bytes, content_type: str):
        self.filename = filename
        self.content_type = content_type or "application/octet-stream"
        self._content = content
        self._stream = BytesIO(content)

    def read(self, size: int = -1) -> bytes:
        if size is None or size < 0:
            return self._content
        return self._content[:size]

    def seek(self, pos: int, whence: int = 0) -> int:
        return self._stream.seek(pos, whence)

    def save(self, path: str) -> None:
        with open(path, "wb") as f:
            self.seek(0)
            f.write(self._stream.read())

    @property
    def size(self) -> int:
        cur = self._stream.tell()
        self._stream.seek(0, 2)
        sz = self._stream.tell()
        self._stream.seek(cur)
        return sz

    @property
    def stream(self) -> BytesIO:
        return self._stream

    def get_stream(self) -> BytesIO:
        return BytesIO(self._content)

    def __repr__(self) -> str:
        return (
            f"<UploadedFile filename={self.filename!r} "
            f"size={self.size} content_type={self.content_type!r}>"
        )

    @staticmethod
    def _from_mapping(data: Mapping[str, Any]) -> "UploadedFile":
        if "filename" not in data:
            raise ValueError("UploadedFile requires 'filename'")

        filename = data["filename"]
        content_type = data.get("content_type", "application/octet-stream")

        if "content" not in data:
            raise ValueError("UploadedFile requires 'content' (bytes)")

        content = data["content"]

        if isinstance(content, (bytes, bytearray)):
            content_bytes = bytes(content)
        elif isinstance(content, memoryview):
            content_bytes = content.tobytes()
        elif isinstance(content, str):
            try:
                content_bytes = base64.b64decode(content, validate=True)
            except Exception as e:
                raise ValueError("Invalid base64 for UploadedFile.content") from e
        else:
            raise TypeError(
                "UploadedFile.content must be bytes/bytearray/memoryview or base64 str"
            )

        if not isinstance(filename, str):
            raise TypeError("UploadedFile.filename must be a string")
        if not isinstance(content_type, str):
            raise TypeError("UploadedFile.content_type must be a string")

        return UploadedFile(
            filename=filename, content=content_bytes, content_type=content_type
        )

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: Any
    ) -> core_schema.CoreSchema:
        def validate(v: Any) -> "UploadedFile":
            if isinstance(v, UploadedFile):
                return v

            if isinstance(v, Mapping):
                return cls._from_mapping(v)

            if isinstance(v, tuple):
                if len(v) == 2:
                    filename, content = v
                    return cls._from_mapping({"filename": filename, "content": content})
                if len(v) == 3:
                    filename, content, content_type = v
                    return cls._from_mapping(
                        {
                            "filename": filename,
                            "content": content,
                            "content_type": content_type,
                        }
                    )
                raise ValueError(
                    "UploadedFile tuple input must be (filename, content) or (filename, content, content_type)"
                )

            raise TypeError("Value is not a valid UploadedFile input")

        def serialize(v: "UploadedFile") -> dict[str, Any]:
            return {
                "filename": v.filename,
                "content_type": v.content_type,
                "size": v.size,
            }

        return core_schema.no_info_plain_validator_function(
            validate,
            serialization=core_schema.plain_serializer_function_ser_schema(
                serialize,
                info_arg=False,
                return_schema=core_schema.dict_schema(),
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: core_schema.CoreSchema, handler: Any
    ) -> dict[str, Any]:
        return {
            "type": "object",
            "title": "UploadedFile",
            "properties": {
                "filename": {"type": "string"},
                "content_type": {"type": "string"},
                "size": {"type": "integer"},
            },
            "required": ["filename"],
            "additionalProperties": True,
            "description": "In-memory uploaded file (serialized as metadata by default).",
        }


def _normalize_multipart_part_headers(raw: bytes) -> bytes:
    """
    Normalize multipart part header names to canonical casing within each part header block.
    """
    out = bytearray()
    i = 0
    n = len(raw)

    while i < n:
        hdr_end = raw.find(b"\r\n\r\n", i)
        if hdr_end == -1:
            out.extend(raw[i:])
            break

        header_block = raw[i:hdr_end]
        rest_start = hdr_end + 4

        if b"\r\n" in header_block and b":" in header_block:
            lines = header_block.split(b"\r\n")
            new_lines = [lines[0]]

            for line in lines[1:]:
                colon = line.find(b":")
                if colon == -1:
                    new_lines.append(line)
                    continue

                name = line[:colon].strip()
                value = line[colon + 1 :].lstrip()

                lname = name.lower()
                if lname == b"content-disposition":
                    cname = b"Content-Disposition"
                elif lname == b"content-type":
                    cname = b"Content-Type"
                elif lname == b"content-length":
                    cname = b"Content-Length"
                elif lname == b"content-transfer-encoding":
                    cname = b"Content-Transfer-Encoding"
                else:
                    # generic Title-Case fallback
                    cname = b"-".join(
                        p[:1].upper() + p[1:].lower() for p in lname.split(b"-")
                    )

                new_lines.append(cname + b": " + value)

            out.extend(b"\r\n".join(new_lines))
            out.extend(b"\r\n\r\n")
            out.extend(raw[rest_start:rest_start])
        else:
            out.extend(raw[i:rest_start])

        i = rest_start
        next_boundary = raw.find(b"\r\n--", i)
        if next_boundary == -1:
            out.extend(raw[i:])
            break
        out.extend(raw[i : next_boundary + 2])  # include the leading \r\n
        i = next_boundary + 2  # position at '--...'
    return bytes(out)


class Request:
    """
    ASGI-style request adapter that lazy-parses JSON, form, and multipart.
    """

    def __init__(
        self,
        scope: dict,
        receive: Callable[..., Any],
        app: "Patera",
        route_parameters: dict | Mapping,
        route_handler: Callable,
    ):
        self._app = app
        self.scope = scope
        self._receive = receive
        self._send: Callable | None = None

        self._body: Union[bytes, None] = None
        self._json: Union[dict, None] = None
        self._form: Union[dict, None] = None
        self._files: Union[dict, None] = None

        self._user: Any = None
        self._route_parameters = route_parameters
        self._route_handler = route_handler
        self._response: Response[Any] = app.response_class(app, self)
        self._context: dict[str, Any] = {}

    @property
    def route_handler(self) -> Callable:
        return self._route_handler

    @property
    def route_parameters(self) -> dict | Mapping:
        return self._route_parameters

    @route_parameters.setter
    def route_parameters(self, rp: dict | Mapping) -> None:
        self._route_parameters = rp

    @property
    def method(self) -> str:
        if self._send is not None:
            return "SOCKET"
        return self.scope.get("method", "").upper()

    @property
    def path(self) -> str:
        return self.scope.get("path", "/")

    @property
    def query_string(self) -> str:
        return self.scope.get("query_string", b"").decode("utf-8")

    @property
    def headers(self) -> dict[str, str]:
        raw = self.scope.get("headers", [])
        return {key.decode("latin1").lower(): val.decode("latin1") for key, val in raw}

    @property
    def query_params(self) -> dict[str, str]:
        qs = self.scope.get("query_string", b"")
        parsed = parse_qs(qs.decode("utf-8"))
        return cast(
            dict[str, str], {k: v if len(v) > 1 else v[0] for k, v in parsed.items()}
        )

    @property
    def user(self) -> Any:
        return self._user

    @property
    def app(self) -> "Patera":
        return self._app

    def set_user(self, user: Any) -> None:
        self._user = user

    def remove_user(self) -> None:
        self._user = None

    async def body(self) -> bytes:
        if self._body is not None:
            return self._body

        parts: list[bytes] = []
        while True:
            msg = await self._receive()
            if msg["type"] == "http.request":
                parts.append(msg.get("body", b""))
                if not msg.get("more_body", False):
                    break
        self._body = b"".join(parts)
        return self._body

    async def json(self) -> dict[str, Any] | None:
        if self._json is not None:
            return self._json
        raw = await self.body()
        if not raw:
            return None
        try:
            self._json = json.loads(raw)
        except json.JSONDecodeError:
            self._json = None
        return self._json

    async def form(self) -> dict[str, Any]:
        if self._form is not None:
            return self._form

        ct = self.headers.get("content-type", "")
        if "multipart/form-data" in ct:
            self._form, self._files = await self._parse_multipart(ct)
        elif "application/x-www-form-urlencoded" in ct:
            raw = await self.body()
            parsed = parse_qs(raw.decode("utf-8"))
            self._form = {k: v if len(v) > 1 else v[0] for k, v in parsed.items()}
            self._files = {}
        else:
            self._form = {}
            self._files = {}

        return self._form

    async def files(self) -> dict[str, UploadedFile]:
        if self._files is None:
            await self.form()
        return self._files or {}

    async def form_and_files(self) -> dict[str, Any]:
        f = await self.form()
        fs = await self.files()
        return {**f, **fs}

    async def send(self, message: dict) -> None:
        if self._send is None:
            raise RuntimeError("Send function is available only on websocket requests")
        return await self._send(message)

    def set_send(self, send: Callable) -> None:
        self._send = send

    async def receive(self) -> dict:
        if self._send is None:
            raise RuntimeError(
                "Receive function is available only on websocket requests"
            )
        return await self._receive()

    async def accept(self) -> None:
        if self._send is None:
            raise RuntimeError(
                "Accept function is available only on websocket requests"
            )
        await self._send({"type": "websocket.accept"})

    @staticmethod
    def _multipart_headers(content_type: str) -> dict[str, str]:
        return {
            "Content-Type": content_type,
            "content-type": content_type,  # be tolerant
        }

    @staticmethod
    def _get_header_case_insensitive(hdrs: Any, key: str) -> str | None:
        if not isinstance(hdrs, dict):
            return None
        candidates: list[Any] = [
            key,
            key.lower(),
            key.upper(),
            key.encode("latin1"),
            key.lower().encode("latin1"),
            key.upper().encode("latin1"),
        ]
        for ck in candidates:
            v = hdrs.get(ck)
            if v is None:
                continue
            if isinstance(v, bytes):
                return v.decode("latin1")
            if isinstance(v, str):
                return v
        return None

    async def _parse_multipart(self, content_type: str) -> tuple[dict, dict]:
        raw = await self.body()
        raw = _normalize_multipart_part_headers(raw)

        stream = BytesIO(raw)

        form_data: dict[str, str] = {}
        files: dict[str, UploadedFile] = {}

        def on_field(field: Any) -> None:
            name = getattr(field, "field_name", None)
            if name is None:
                return
            val = getattr(field, "value", None)
            if isinstance(name, bytes):
                name = name.decode("latin1")
            if isinstance(val, bytes):
                val = val.decode("utf-8", "replace")
            form_data[name] = cast(str, val)

        def on_file(f: Any) -> None:
            raw_name = getattr(f, "field_name", None)
            if raw_name is None:
                return
            raw_fn = getattr(f, "file_name", b"") or b""
            name = (
                raw_name.decode("latin1") if isinstance(raw_name, bytes) else raw_name
            )
            fn = raw_fn.decode("latin1") if isinstance(raw_fn, bytes) else raw_fn

            fileobj = f.file_object
            fileobj.seek(0)
            content = fileobj.read()

            hdrs = getattr(f, "headers", {}) or {}
            part_ct = (
                self._get_header_case_insensitive(hdrs, "Content-Type")
                or "application/octet-stream"
            )

            files[name] = UploadedFile(
                filename=fn, content=content, content_type=part_ct
            )

        pm.parse_form(
            headers={"Content-Type": content_type, "content-type": content_type},
            input_stream=stream,
            on_field=on_field,
            on_file=on_file,
        )

        return form_data, files

    async def get_data(self, location: str = "json") -> dict[str, Any] | None:
        if location == "json":
            return await self.json()
        if location == "form":
            return await self.form()
        if location == "files":
            return await self.files()
        if location == "form_and_files":
            return await self.form_and_files()
        if location == "query":
            return self.query_params
        return None

    @property
    def response(self) -> Response:
        return self._response

    @property
    def res(self) -> Response:
        return self._response

    @property
    def context(self) -> dict[str, Any]:
        return self._context



"""
Utility methods for controller related things
"""

from typing import (
    Any,
    Annotated,
    Type,
    Optional,
    get_origin,
    get_args,
    Mapping,
    get_type_hints,
)
import inspect
from pydantic import BaseModel

from ..request import Request
from ..response import Response
from ..media_types import MediaType


def _get_handler_dict(obj: Any) -> dict[str, Any]:
    """Return or create the _handler dict on an object."""
    d = getattr(obj, "_handler", None)
    if not isinstance(d, dict):
        d = {}
        setattr(obj, "_handler", d)
    return d


def _unwrap_annotated(tp: Any) -> Any:
    """If Annotated[T, ...], return T; else the original."""
    if get_origin(tp) is Annotated:
        args = get_args(tp)
        return args[0] if args else tp
    return tp


def _is_subclass(x: Any, base: Type) -> bool:
    try:
        return inspect.isclass(x) and issubclass(x, base)
    except Exception:
        return False


def _is_pydantic_model(tp: Any) -> bool:
    return _is_subclass(tp, BaseModel)


def _build_model(model_cls: Type[BaseModel], data: Any) -> BaseModel:
    return model_cls.model_validate(data)  # type: ignore[attr-defined]


def _content_type_matches(incoming: str, expected: "MediaType") -> bool:
    """
    Accept parameters (e.g., '; charset=utf-8') and handle +json suffixes.
    """
    inc = (incoming or "").lower()
    exp = expected.value.lower()
    base = inc.split(";")[0].strip()

    if exp == MediaType.APPLICATION_JSON.value:
        # RFC 6839: */*+json is valid too
        return base == "application/json" or base.endswith("+json")
    if exp == MediaType.APPLICATION_PROBLEM_JSON.value:
        # RFC 7807; accept application/problem+json and */*+json
        return base in (
            "application/problem+json",
            "application/json",
        ) or base.endswith("+json")
    if exp == MediaType.MULTIPART_FORM_DATA.value:
        return base == "multipart/form-data"
    if exp == MediaType.APPLICATION_X_WWW_FORM_URLENCODED.value:
        return base == "application/x-www-form-urlencoded"
    return base == exp


async def _read_payload_for_consumes(
    req: "Request", mt: "MediaType"
) -> Mapping[str, Any]:
    """
    Map declared MediaType → Request loader.
    Returns mapping suitable for building Pydantic models.
    """
    if mt in (MediaType.APPLICATION_JSON, MediaType.APPLICATION_PROBLEM_JSON):
        return (await req.json()) or {}
    if mt == MediaType.APPLICATION_X_WWW_FORM_URLENCODED:
        return await req.form()
    if mt == MediaType.MULTIPART_FORM_DATA:
        return await req.form_and_files()
    # extend with additional types if needed.
    return {}


def _extract_response_type(func) -> Optional[Type[Any]]:
    """
    If the function is annotated as -> Response[T], return T; else None.
    """
    hints = get_type_hints(func, include_extras=True)
    ret = hints.get("return")
    if ret is None:
        return None
    if get_origin(ret) is Response:
        args = get_args(ret)
        if args:
            t = args[0]
            if get_origin(t) is Annotated:  # peel Annotated[T, ...]
                t = get_args(t)[0]
            return t
    return None



"""
Patera cli
"""

import argparse
from typing import Callable
from pathlib import Path

from .start_project import start_dev, start_prod, start_cli, start_testing

methods: dict[str, Callable] = {
    "dev": start_dev,
    "prod": start_prod,
    "cli": start_cli,
    "test": start_testing,
}


def main():
    parser = argparse.ArgumentParser(prog="patera")
    subparsers = parser.add_subparsers(dest="command")

    start_dev_parser = subparsers.add_parser("dev")
    start_dev_parser.add_argument(
        "--app",
        type=str,
        default=None,
        required=False,
        help="Import string of App class implementation",
    )
    start_dev_parser.add_argument(
        "--env-file",
        type=str,
        default=None,
        required=False,
        help="Environment file to load",
    )

    start_prod_parser = subparsers.add_parser("prod")
    start_prod_parser.add_argument(
        "--app",
        type=str,
        default=None,
        required=False,
        help="Import string of App class implementation",
    )
    start_prod_parser.add_argument(
        "--env-file",
        type=str,
        default=None,
        required=False,
        help="Environment file to load",
    )

    start_cli_parser = subparsers.add_parser("cli")
    start_cli_parser.add_argument(
        "command_name",
        type=str,
        help="CLI command name to run",
    )
    start_cli_parser.add_argument(
        "command_args",
        nargs=argparse.REMAINDER,
        help="Arguments passed to the CLI command",
    )

    start_testing_parser = subparsers.add_parser("test")
    start_testing_parser.add_argument(
        "pytest_args",
        nargs=argparse.REMAINDER,
        help="Arguments passed to pytest",
    )

    args = parser.parse_args()

    method = methods.get(args.command)
    if method is None:
        parser.print_help()
        return

    args_dict = vars(args)

    return method(Path.cwd(), **args_dict)



"""
OpenAPI controller
"""

import re
from typing import Any, Dict, Optional, Set, Type, Tuple, List, cast, Callable
from pydantic import BaseModel

from .http_statuses import HttpStatus
from .media_types import MediaType
from .response import Response
from .request import Request
from .controller import Controller, get, produces
from .controller.utilities import _extract_response_type

_WERKZEUG_PARAM_RE = re.compile(r"<(?:(int|string|path):)?([a-zA-Z_][a-zA-Z0-9_]*)>")


class OpenApiSpecs(BaseModel):
    openapi: str = "3.0.3"
    info: Any = None
    servers: Any = None
    paths: Any = None
    components: Any = None


class OpenAPIController(Controller):
    @get("/specs.json", open_api_spec=False)
    @produces(MediaType.APPLICATION_JSON)
    async def specs(self, req: Request) -> Response[OpenApiSpecs]:
        return req.response.json(self.app.json_spec).status(HttpStatus.OK)

    @get("/docs", open_api_spec=False)
    @produces(MediaType.TEXT_HTML)
    async def docs(self, req: Request) -> Response:
        return (
            await req.response.html_from_string("""
        <!DOCTYPE html>
        <html>
            <head>
                <title>Swagger UI</title>
                <link rel="stylesheet"
                      href="https://cdn.jsdelivr.net/npm/swagger-ui-dist@4.18.3/swagger-ui.css" />
            </head>
            <body>
                <div id="swagger-ui"></div>
                <script src="https://cdn.jsdelivr.net/npm/swagger-ui-dist@4.18.3/swagger-ui-bundle.js"></script>
                <script>
                    const ui = SwaggerUIBundle({
                        url: '{{ url_for("OpenAPIController.specs") }}',
                        dom_id: '#swagger-ui',
                    })
                </script>
            </body>
        </html>
        """)
        ).status(HttpStatus.OK)


def _as_media_type(mt: Any) -> str:
    # Accept enum-ish (with .value) or plain str
    if mt is None:
        return "application/json"
    return getattr(mt, "value", str(mt))


def _as_status_code(http_status: HttpStatus | int) -> int:
    if isinstance(http_status, HttpStatus):
        return http_status.value
    return http_status


def _model_name(model: Type) -> str:
    # Pick a stable schema key
    return getattr(model, "__name__", "AnonymousModel")


def _pydantic_schema_dict(model: Optional[Type]) -> Optional[Dict[str, Any]]:
    if model is None:
        return None
    # Pydantic v2
    if hasattr(model, "model_json_schema"):
        try:
            return model.model_json_schema(ref_template="#/components/schemas/{model}")
        except TypeError:
            # older/stricter signatures
            return model.model_json_schema()
    # Pydantic v1
    if hasattr(model, "schema"):
        try:
            return model.schema(ref_template="#/components/schemas/{model}")
        except TypeError:
            return model.schema()
    # Fallback – unknown object
    return {"type": "object"}


def _rebuild_model(model: Optional[Type]) -> None:
    if model is None:
        return
    # Pydantic v2
    if hasattr(model, "model_rebuild"):
        try:
            model.model_rebuild()
        except TypeError:
            pass
    # Pydantic v1
    if hasattr(model, "update_forward_refs"):
        try:
            # model.update_forward_refs()
            model.model_rebuild()
        except TypeError:
            pass


def _pydantic_schema_and_defs(
    model: Optional[Type],
) -> Tuple[Optional[Dict[str, Any]], Dict[str, Dict[str, Any]]]:
    """
    Returns (schema_for_model, defs) where defs are nested component schemas
    that must be merged into components/schemas.
    """
    if model is None:
        return None, {}

    _rebuild_model(model)

    # Pydantic v2
    if hasattr(model, "model_json_schema"):
        try:
            schema = model.model_json_schema(
                ref_template="#/components/schemas/{model}"
            )
        except TypeError:
            schema = model.model_json_schema()
        defs = schema.pop("$defs", {}) or {}
        return schema, defs

    # Pydantic v1
    if hasattr(model, "schema"):
        try:
            schema = model.schema(ref_template="#/components/schemas/{model}")
        except TypeError:
            schema = model.schema()
        defs = schema.pop("definitions", {}) or {}
        return schema, defs

    # Fallback
    return {"type": "object"}, {}


def _ensure_schema(
    components: Dict[str, Any], model: Optional[Type]
) -> Optional[Dict[str, Any]]:
    if model is None:
        return None

    name = _model_name(model)
    if "schemas" not in components:
        components["schemas"] = {}

    # Get schema and nested defs
    schema, defs = _pydantic_schema_and_defs(model)

    # Merge nested defs first (don't overwrite if already present)
    for def_name, def_schema in defs.items():
        components["schemas"].setdefault(def_name, def_schema)

    # Register the model schema itself (if not already present)
    if name not in components["schemas"]:
        components["schemas"][name] = schema or {"type": "object"}

    # Return a $ref to the component
    return {"$ref": f"#/components/schemas/{name}"}


_TYPE_MAP = {
    "int": {"type": "integer", "format": "int64"},
    "string": {"type": "string"},
    "path": {
        "type": "string"
    },  # openapi does not have a special type for path. Will be treated as string with x-greedy-path: true
}


def _convert_path_and_extract_params(path: str) -> Tuple[str, List[Dict[str, Any]]]:
    """
    Convert Flask-style route params in `path` to OpenAPI templating and
    return (converted_path, parameters_list).
    """
    params: List[Dict[str, Any]] = []
    seen: Set[str] = set()

    def _repl(m: re.Match) -> str:
        typ = m.group(1) or "string"
        name = m.group(2)
        if name not in seen:
            seen.add(name)
            schema = _TYPE_MAP.get(typ, {"type": "string"})
            param: Dict[str, Any] = {
                "in": "path",
                "name": name,
                "required": True,
                "schema": dict(schema),  # copy
            }
            if typ == "path":
                param["x-greedy-path"] = True
                param["description"] = (param.get("description") or "") + (
                    " Greedy path segment captured by the server."
                )
            params.append(param)
        return "{" + name + "}"

    converted = _WERKZEUG_PARAM_RE.sub(_repl, path)
    return converted, params


def build_openapi(
    controllers: Dict[str, "Controller"],
    *,
    title: str,
    version: str,
    openapi_version: str,
    servers: Optional[list[str]] = None,
) -> Dict[str, Any]:
    """
    controllers structure: {
        "<base_path>": {
            "<HTTP_METHOD>": {
                "<endpoint_path>": {
                    "method": <callable>,
                    "produces": <MediaType>,
                    "consumes": <MediaType>,
                    "consumes_type": <PydanticModel>,
                    "response_type": <PydanticModel>, # optional
                    "default_status_code": <HttpStatus>,
                    "http_method": <HTTP_METHOD>,
                    "path": <endpoint_path>,
                    "base_path": <base_path>,
                    "summary": <string>, # optional
                    "description": <string>, # optional
                    "tags": [<string>, ...], # optional
                    "operation_id": <string>, # optional
                    "error_responses": [<Descriptor>, ...]
                }
            }
        }
    }
    Descriptor has: media_type, body, description, status (default BAD_REQUEST).
    """
    spec: Dict[str, Any] = {
        "openapi": openapi_version,
        "info": {"title": title, "version": version, "description": ""},
        "paths": {},
        "components": {"schemas": {}},
    }
    if servers:
        spec["servers"] = [{"url": url} for url in servers]

    components = spec["components"]
    paths = spec["paths"]

    referenced_models: Set[Type] = set()

    for base_path, controller in controllers.items():
        if not controller.open_api_spec:
            continue

        for http_method, endpoints_map in controller.endpoints_map.items():
            for endpoint_path, ep_cfg in endpoints_map.items():
                if not cast(dict, ep_cfg).get("open_api_spec"):
                    continue

                # Build raw path first
                raw_full_path = (
                    f"/{base_path.strip('/')}/{endpoint_path.strip('/')}".replace(
                        "//", "/"
                    )
                )
                if base_path == "/":
                    raw_full_path = f"/{endpoint_path.strip('/')}" or "/"

                # Converts dynamic paths to OpenApi complient format and generates path schema
                full_path, path_params = _convert_path_and_extract_params(raw_full_path)

                if full_path not in paths:
                    paths[full_path] = {}

                op_obj: Dict[str, Any] = {}
                paths[full_path][http_method.lower()] = op_obj

                # Summary / description
                op_obj["summary"] = cast(dict, ep_cfg).get("summary", "")
                op_obj["description"] = cast(dict, ep_cfg).get(
                    "method"
                ).__doc__ or cast(dict, ep_cfg).get("description", "")

                # Tags
                tags = cast(dict, ep_cfg).get("tags")
                if tags:
                    op_obj["tags"] = tags

                # operationId
                op_obj["operationId"] = cast(
                    Callable, cast(dict[str, Callable], ep_cfg).get("method")
                ).__name__ or (
                    f"{http_method.lower()}_{base_path.strip('/').replace('/', '_')}_{endpoint_path.strip('/').replace('/', '_')}"
                    or f"{http_method.lower()}_root"
                )

                # Path parameters
                if path_params:
                    op_obj["parameters"] = [*path_params]

                # Request body
                consumes = cast(str, cast(dict, ep_cfg).get("consumes"))
                consumes_type = cast(Type, cast(dict, ep_cfg).get("consumes_type"))
                if consumes or consumes_type:
                    mt = _as_media_type(consumes) if consumes else "application/json"
                    schema_ref = _ensure_schema(components, consumes_type)
                    if consumes_type:
                        try:
                            referenced_models.add(consumes_type)
                        except TypeError:
                            pass
                    op_obj["requestBody"] = {
                        "required": True if consumes_type else False,
                        "content": {mt: {"schema": schema_ref or {"type": "object"}}},
                    }

                # Responses
                produces = cast(str, cast(dict, ep_cfg).get("produces"))
                mt_out = _as_media_type(produces) if produces else "application/json"
                default_status = _as_status_code(
                    cast(int, cast(dict, ep_cfg).get("default_status_code", 200))
                )
                responses: Dict[int, Any] = {}
                op_obj["responses"] = responses

                response_type = _extract_response_type(
                    cast(str, cast(dict, ep_cfg).get("method"))
                )
                resp_schema_ref = (
                    _ensure_schema(components, response_type) if response_type else None
                )
                if response_type:
                    try:
                        referenced_models.add(response_type)
                    except TypeError:
                        pass

                responses[default_status] = {
                    "description": cast(str, cast(dict, ep_cfg).get("description", ""))
                    or "",
                    "content": {
                        mt_out: {"schema": resp_schema_ref or {"type": "object"}}
                    },
                }

                # Error responses
                for desc in (
                    cast(list, cast(dict, ep_cfg).get("error_responses", [])) or []
                ):
                    status_code = _as_status_code(
                        cast(int, getattr(desc, "status", None))
                    )
                    err_response_type = _as_media_type(
                        getattr(desc, "media_type", None)
                    )
                    body_model = getattr(desc, "body", None)
                    err_desc = getattr(desc, "description", None) or ""

                    schema_ref = _ensure_schema(components, body_model)
                    if body_model:
                        try:
                            referenced_models.add(body_model)
                        except TypeError:
                            pass

                    if status_code not in responses:
                        responses[status_code] = {
                            "description": err_desc or "",
                            "content": {},
                        }
                    if not responses[status_code].get("description"):
                        responses[status_code]["description"] = err_desc or ""

                    responses[status_code]["content"][err_response_type] = {
                        "schema": schema_ref or {"type": "object"}
                    }

    return spec



"""
Base extension class
"""

from typing import TYPE_CHECKING, Any, Optional, cast
from abc import abstractmethod, ABC
from pydantic import BaseModel, ValidationError

if TYPE_CHECKING:
    from .patera import Patera


class BaseExtension(ABC):
    _configs_name: str
    _app: "Optional[Patera]"
    _configs: dict[str, Any]

    @abstractmethod
    def init_app(self, app: "Patera") -> None: ...

    def validate_configs(
        self, configs: dict[str, Any], model: type[BaseModel]
    ) -> dict[str, Any]:
        try:
            return model.model_validate(configs).model_dump()
        except ValidationError as e:
            raise ValueError(
                f"Invalid configuration for {self._configs_name or self.__class__.__name__}: {e}"
            ) from e

    @property
    def configs_name(self) -> str:
        """
        Return the config name used in app configurations
        for this extension.
        """
        return self._configs_name

    @property
    def app(self) -> "Patera":
        if self._app is None:
            raise RuntimeError("Extension not initialized with a Patera app.")
        return cast("Patera", self._app)

    @property
    def configs(self) -> dict[str, Any]:
        """Returns a dictinary of extension configs"""
        return self._configs

    @property
    def nice_name(self) -> str | None:
        """Returns nice name of the extension or None"""
        return self._configs.get("NICE_NAME", None)



"""Controller subpackage"""

from .controller import Controller, Descriptor, path
from .decorators import (
    get,
    post,
    delete,
    patch,
    put,
    before_request,
    after_request,
    produces,
    consumes,
    open_api_docs,
    cors,
    no_cors,
    socket,
    development,
)

__all__ = [
    "Controller",
    "path",
    "get",
    "post",
    "put",
    "patch",
    "delete",
    "consumes",
    "produces",
    "Descriptor",
    "open_api_docs",
    "before_request",
    "after_request",
    "cors",
    "no_cors",
    "socket",
    "development",
]



"""
In-memory log buffer
"""

from collections import deque
from threading import Lock
from typing import Deque, Dict, Any, List


class InMemoryLogBuffer:
    """
    Callable Loguru sink that stores recent log records in memory.

    - Uses a bounded deque so memory doesn't grow indefinitely.
    - Thread-safe with a simple Lock.
    """

    def __init__(self, maxlen: int = 1000):
        self._buffer: Deque[Dict[str, Any]] = deque(maxlen=maxlen)
        self._severe_buffer: Deque[Dict[str, Any]] = deque(maxlen=maxlen)
        self._lock = Lock()

    def __call__(self, message):  # type: ignore[override]
        record = message.record
        data = {
            "time": record["time"].isoformat(),
            "level": record["level"].name,
            "message": record["message"],
            "logger_name": record["extra"].get("logger_name"),
            "file": record["file"].name,
            "line": record["line"],
            "name": record["name"],
            "function": record["function"],
        }

        with self._lock:
            self._buffer.append(data)

        if record["level"].name in ["ERROR", "CRITICAL", "FATAL"]:
            with self._lock:
                self._severe_buffer.append(data)

    def get_last(self, n: int = 100) -> List[Dict[str, Any]]:
        """Return last n records (newest last)."""
        with self._lock:
            if n >= len(self._buffer):
                return list(self._buffer)
            return list(self._buffer)[-n:]

    def get_all(self) -> List[Dict[str, Any]]:
        with self._lock:
            return list(self._buffer)

    def get_last_severe(self, n: int = 100) -> List[Dict[str, Any]]:
        """Return last n records (newest last)."""
        with self._lock:
            if n >= len(self._severe_buffer):
                return list(self._severe_buffer)
            return list(self._severe_buffer)[-n:]

    def get_severe(self) -> List[Dict[str, Any]]:
        with self._lock:
            return list(self._severe_buffer)



"""
Collection of http exceptions that can be raised
"""

from typing import Any, Optional
from ..http_statuses import HttpStatus


class BaseHttpException(Exception):
    """
    Base http exception class
    """

    def __init__(
        self,
        message: str = "",
        status_code: int | HttpStatus = 500,
        status="error",
        data=None,
    ):
        """
        Init method
        """
        self.message = message
        if isinstance(status_code, HttpStatus):
            status_code = status_code.value
        self.status_code = status_code
        self.status = status
        self.data = data


class StaticAssetNotFound(BaseHttpException):
    """
    HTTP exception for static assets not found
    """

    def __init__(
        self,
        message: str = "Static asset not found",
        status_code: int = 404,
        status: str = "error",
        data: Any = None,
    ):
        super().__init__(message, status_code, status, data)


class AborterException(BaseHttpException):
    """
    Aborter exception
    """

    def __init__(
        self,
        message: str = "",
        status_code: int = 400,
        status: str = "error",
        data: Optional[Any] = None,
    ):
        super().__init__(message, status_code, status, data)


class HtmlAborterException(BaseHttpException):
    """
    Html aborter exception
    """

    def __init__(
        self, template: str, status_code: int | HttpStatus, data: Optional[Any] = None
    ):
        super().__init__("Error", status_code=status_code, data=data)
        self.template = template
        if isinstance(status_code, HttpStatus):
            status_code = status_code.value
        self.status_code = status_code
        self.data = data


def abort(
    msg: str,
    status_code: int | HttpStatus = HttpStatus.BAD_REQUEST,
    status: str = "error",
    data: Any = None,
):
    """
    Aborts request by raising an aborter exception
    """
    if isinstance(status_code, HttpStatus):
        status_code = status_code.value
    raise AborterException(msg, status_code, status, data)


def html_abort(
    template: str,
    status_code: int | HttpStatus = HttpStatus.BAD_REQUEST,
    data: Any = None,
):
    """
    Aborts request with html response
    """
    if isinstance(status_code, HttpStatus):
        status_code = status_code.value
    raise HtmlAborterException(template, status_code, data)



"""
Custom exception classes for Patera
"""


class CustomException(Exception):
    """Base custom exception class"""


class MethodNotControllerMethod(CustomException):
    """
    Error if the decorated method is not part of a controller
    """

    def __init__(self, message):
        self.message = message


class UnexpectedDecorator(CustomException):
    """
    Error for unexpected decorators
    """

    def __init__(self, message):
        self.message = message



"""
Base logger configuration class
"""

import sys
import re
from abc import ABC
from datetime import timedelta
from pathlib import Path
from typing import (
    Dict,
    Any,
    Union,
    Optional,
    ClassVar,
    cast,
    TYPE_CHECKING,
    IO,
    Callable,
    Protocol,
    runtime_checkable,
    Type,
    TypedDict,
    NotRequired,
)
from enum import StrEnum
from pydantic import BaseModel, field_validator, Field, ConfigDict

from loguru import logger

if TYPE_CHECKING:
    from ..patera import Patera


@runtime_checkable
# pylint: disable=R0903
class Writable(Protocol):
    def write(self, s: str) -> Any: ...


SinkType = Union[str, Path, IO[str], IO[bytes], Callable[[str], Any]]
SinkInput = Union[str, Path, IO[str], IO[bytes], Callable[[str], Any], Writable]
SinkAccepted = Union[str, Writable, Callable[[str], None]]
RotationType = Union[str, int, timedelta, Callable[[str, Any], Any]]
RetentionType = Union[str, int, timedelta, Callable[[str, Any], Any]]
CompressionType = Optional[str]
FilterType = Union[None, str, Dict[str, str], Callable[[Dict[str, Any]], bool]]


class OutputSink(StrEnum):
    STDERR = "STDERR"
    STDOUT = "STDOUT"


class LogLevel(StrEnum):
    TRACE = "TRACE"
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class _LoggerConfigBase(BaseModel):
    """
    Base config model for loggers
    """

    model_config = ConfigDict(extra="allow")

    SINK: Optional[str | Path] = Field(
        OutputSink.STDERR.value, description="Output of the logger"
    )
    LEVEL: Optional[LogLevel] = Field(
        LogLevel.TRACE, description="Log level for logger implementation"
    )
    FORMAT: Optional[str] = Field(
        "<green>{time:HH:mm:ss}</green> | <level>{level}</level> | {extra[logger_name]} | <level>{message}</level>",
        description="Output format",
    )
    ENQUEUE: Optional[bool] = Field(
        False,
        description="Use enqueue or not. For thread, process and coroutine safe logging.",
    )
    BACKTRACE: Optional[bool] = Field(True, description="Use backtrace")
    DIAGNOSE: Optional[bool] = Field(True, description="Diagnose mode")
    COLORIZE: Optional[bool] = Field(True, description="Colorize output")
    DELAY: Optional[bool] = Field(True, description="Use delay")
    ROTATION: Optional[RotationType] = Field(
        "5 MB", description="Size of log file before rotating."
    )
    RETENTION: Optional[RetentionType] = Field(
        "10 days", description="How long log files are kept before automatic erasure."
    )
    COMPRESSION: CompressionType = Field(
        "zip", description="If log files should be compressed. No compression = None"
    )
    SERIALIZE: Optional[bool] = Field(False)
    ENCODING: Optional[str] = Field(
        "utf-8", description="Encoding of logged messages - for file logging."
    )
    MODE: Optional[str] = Field(
        "a", description="If log messages should be appended to the file."
    )

    @field_validator("SINK", mode="before")
    @classmethod
    def normalize_sink(cls, v):
        """Accept case-insensitive strings and convert to canonical form."""
        s = v.strip().upper()
        if s in {"STDERR", "SYS.ERR", "CONSOLE", "ERR"}:
            return "STDERR"
        if s in {"STDOUT", "SYS.OUT", "OUT"}:
            return "STDOUT"
        if s in {"NULL", "DEVNULL", "NONE"}:
            return "NULL"
        # Otherwise treat it as a path
        return Path(v)


class LoggerConfig(TypedDict):
    SINK: NotRequired[str | Path]
    LEVEL: NotRequired[LogLevel]
    FORMAT: NotRequired[str]
    ENQUEUE: NotRequired[bool]
    BACKTRACE: NotRequired[bool]
    DIAGNOSE: NotRequired[bool]
    COLORIZE: NotRequired[bool]
    DELAY: NotRequired[bool]
    ROTATION: NotRequired[RotationType]
    RETENTION: NotRequired[RetentionType]
    COMPRESSION: NotRequired[CompressionType]
    SERIALIZE: NotRequired[bool]
    ENCODING: NotRequired[str]
    MODE: NotRequired[str]


class LoggerBase(ABC):
    """
    Base template for configuring a Loguru sink.
    Subclasses implement get_sink() and optionally override getters below.
    """

    configs_model: ClassVar[Optional[Type[_LoggerConfigBase]]] = cast(
        Type[_LoggerConfigBase], None
    )

    def __init__(self, app: "Patera"):
        self.app: "Patera" = app
        # loads configs for the logger from application configurations
        # by the config class name as upper-case
        # example: CustomLoggerConfig -> CUSTOM_LOGGER_CONFIG
        self.conf: Dict[str, Any] = app.get_conf(self.logger_name, None) or {}  # type: ignore
        self.conf = self.validate_configs(self.conf)

    def validate_configs(self, configs: dict[str, Any]) -> dict[str, Any]:
        if self.configs_model is not None and issubclass(
            self.configs_model, _LoggerConfigBase
        ):
            return self.configs_model.model_validate(configs).model_dump()
        return _LoggerConfigBase.model_validate(configs).model_dump()

    @property
    def logger_name(self) -> str:
        """Returns class name as upper snake case"""
        name = self.__class__.__name__
        return re.sub(r"(?<!^)(?=[A-Z])", "_", name).upper()

    def get_sink(self) -> SinkInput:
        sink = self.conf.get("SINK", "STDERR")
        if sink == "STDERR":
            return sys.stderr
        if sink == "STDOUT":
            return sys.stdout
        if sink == "NULL":
            return "NUL" if sys.platform.startswith("win") else "/dev/null"
        p = Path(sink)
        p.parent.mkdir(parents=True, exist_ok=True)
        return p

    def get_conf_value(self, key: str, default: Any = None) -> Any:
        return self.conf.get(key, default)

    def get_level(self) -> Union[str, int]:
        return self.get_conf_value("LEVEL", LogLevel.INFO)

    def get_format(self) -> str:
        # includes the logger name as a constant-like tag
        return self.get_conf_value(
            "FORMAT",
            "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
            "<level>{level: <8}</level> | {extra[logger_name]} | "
            "{name}:{function}:{line} - <cyan>{message}</cyan>",
        )

    def dict_to_format_string(self, fmt_dict: dict[str, str]) -> str:
        """
        Converts a dictionary of Loguru placeholders into a JSON-like
        format string suitable for loguru.add(format=...).

        CURRENTLY DOESN'T WORK AS INTENDED
        """
        items = []
        for key, value in fmt_dict.items():
            items.append(f'"{key}": "{value}"')
        return "{{{}}}".format(", ".join(items))

    def get_rotation(self) -> Optional[RotationType]:
        return self.get_conf_value("ROTATION", None)

    def get_retention(self) -> Optional[RetentionType]:
        return self.get_conf_value("RETENTION", None)

    def get_compression(self) -> CompressionType:
        return self.get_conf_value("COMPRESSION", None)

    def get_filter(self) -> FilterType:
        return None

    def _wrap_filter_with_logger_name(
        self, original_filter: FilterType
    ) -> Callable[[Dict[str, Any]], bool]:
        def _wrapped(record: Dict[str, Any]) -> bool:
            # Injects the logger name as extra information
            record["extra"].setdefault("logger_name", self.logger_name)

            if original_filter is None:
                return True
            if callable(original_filter):
                return bool(original_filter(record))
            if isinstance(original_filter, str):
                return record.get("name") == original_filter
            if isinstance(original_filter, dict):
                return all(
                    record["extra"].get(k) == v for k, v in original_filter.items()
                )
            return True

        return _wrapped

    def get_enqueue(self) -> bool:
        return self.get_conf_value("ENQUEUE", True)

    def get_backtrace(self) -> bool:
        return self.get_conf_value("BACKTRACE", False)

    def get_diagnose(self) -> bool:
        return self.get_conf_value("DIAGNOSE", False)

    def get_colorize(self) -> Optional[bool]:
        return self.get_conf_value("COLORIZE", None)

    def get_serialize(self) -> bool:
        return self.get_conf_value("SERIALIZE", False)

    def get_encoding(self) -> Optional[str]:
        return self.get_conf_value("ENCODING", "utf-8")

    def get_mode(self) -> Optional[str]:
        return self.get_conf_value("MODE", "a")

    def get_delay(self) -> bool:
        return self.get_conf_value("DELAY", True)

    def remove_existing_handlers(self) -> bool:
        return self.get_conf_value("remove_existing_handlers", False)

    def _build_handler_kwargs(self) -> Dict[str, Any]:
        return {
            "level": self.get_level(),
            "format": self.get_format(),
            "rotation": self.get_rotation(),
            "retention": self.get_retention(),
            "compression": self.get_compression(),
            "filter": self.get_filter(),
            "enqueue": self.get_enqueue(),
            "backtrace": self.get_backtrace(),
            "diagnose": self.get_diagnose(),
            "colorize": self.get_colorize(),
            "serialize": self.get_serialize(),
            "encoding": self.get_encoding(),
            "mode": self.get_mode(),
            "delay": self.get_delay(),
        }

    def _is_path_sink(self, sink: SinkAccepted) -> bool:
        return isinstance(sink, str)

    def _is_stream_sink(self, sink: SinkAccepted) -> bool:
        return (not isinstance(sink, str)) and hasattr(sink, "write")

    def _is_callable_sink(self, sink: SinkAccepted) -> bool:
        return callable(sink) and not hasattr(sink, "write")

    def _normalize_sink(self, sink: SinkInput) -> SinkAccepted:
        if isinstance(sink, Path):
            return sink.as_posix()
        return sink  # type: ignore[return-value]

    def _filter_kwargs_for_sink(
        self, sink: SinkAccepted, kwargs: Dict[str, Any]
    ) -> Dict[str, Any]:
        # Common kwargs accepted by all sinks
        common = {
            "level",
            "format",
            "filter",
            "colorize",
            "serialize",
            "backtrace",
            "diagnose",
            "enqueue",
            "catch",
        }
        # File-path only kwargs (Loguru opens the file for you)
        file_only = {
            "rotation",
            "retention",
            "compression",
            "mode",
            "delay",
            "encoding",
            "buffering",
        }
        # Streams (TextIO / Writable): no rotation/retention/compression/encoding/mode/delay
        stream_only: set = set()  # (no extra)
        # Callables: also no file-only kwargs
        callable_only: set = set()

        if self._is_path_sink(sink):
            allowed = common | file_only
        elif self._is_stream_sink(sink):
            allowed = common | stream_only
        else:  # callable sink
            allowed = common | callable_only

        return {k: v for k, v in kwargs.items() if k in allowed and v is not None}

    def configure(self) -> int:

        raw_sink = self.get_sink()
        sink = self._normalize_sink(raw_sink)

        base_kwargs = self._build_handler_kwargs()
        # Wrap filter to inject extra["logger_name"] for every record to this sink
        base_kwargs["filter"] = self._wrap_filter_with_logger_name(
            base_kwargs.get("filter")
        )

        kwargs = self._filter_kwargs_for_sink(sink, base_kwargs)
        logger_sink_id: int = logger.add(sink, **kwargs)

        return logger_sink_id



"""
Init file for Patera package
"""

from .patera import Patera, app, app_path, on_shutdown, on_startup
from .base_extension import BaseExtension
from .middleware import MiddlewareBase
from .exceptions import abort, html_abort
from .configuration_base import BaseConfig

from .request import Request, UploadedFile
from .response import Response

from .utilities import run_sync_or_async, run_in_background
from .media_types import MediaType
from .http_methods import HttpMethod
from .http_statuses import HttpStatus
from .logging.logger_config_base import LogLevel

__all__ = [
    "Patera",
    "abort",
    "Request",
    "Response",
    "run_sync_or_async",
    "run_in_background",
    "UploadedFile",
    "MediaType",
    "HttpMethod",
    "HttpStatus",
    "html_abort",
    "app",
    "app_path",
    "on_shutdown",
    "on_startup",
    "BaseExtension",
    "BaseConfig",
    "LogLevel",
    "MiddlewareBase",
]



"""
Start project in dev or prod mode
"""

import os
import sys
import inspect
from pathlib import Path
import importlib
from typing import Optional, Type

from dotenv import load_dotenv

from ..utilities import import_module


def load_env_file(cwd: str, env_file: Optional[str] = None) -> Optional[Path]:
    """
    Load environment variables from a .env file.

    Rules:
    - If env_file is provided, load that file relative to cwd unless absolute.
    - If env_file is not provided, try '.env' and then '.env.dev'.
    - Only check inside cwd for implicit lookup.
    - Returns the loaded file path, or None if no file was loaded.
    """
    root = Path(cwd).resolve()

    if env_file:
        env_path = Path(env_file)
        if not env_path.is_absolute():
            env_path = root / env_path

        if not env_path.is_file():
            raise RuntimeError(f"Environment file not found: {env_path}")

        load_dotenv(dotenv_path=env_path, override=False)
        return env_path

    for name in [".env", ".env.dev", ".env.prod"]:
        env_path = root / name
        if env_path.is_file():
            load_dotenv(dotenv_path=env_path, override=False)
            return env_path

    return None


def path_to_module(file_path: Path, root: Path) -> str:
    """
    Convert a Python file path to an importable module path.

    Example:
        /project/app.py -> app
        /project/api/server.py -> api.server
        /project/app/__init__.py -> app
        /project/app/main.py -> app.main
    """
    rel = file_path.relative_to(root)

    if rel.name == "__init__.py":
        rel = rel.parent
    else:
        rel = rel.with_suffix("")

    return ".".join(rel.parts)


def has_python_files(path: Path) -> bool:
    return path.exists() and path.is_dir() and any(path.glob("*.py"))


def iter_candidate_files(root: Path):
    """
    Yield candidate Python files in priority order.

    Priority:
    1. Common entrypoint names in root/app
    2. Common entrypoint names in root
    3. Common entrypoint names anywhere below root
    4. All remaining Python files
    """
    common_names = [
        "__init__.py",
        "app.py",
        "main.py",
        "server.py",
        "run.py",
        "application.py",
    ]

    yielded: set[Path] = set()
    app_dir = root / "app"

    def add_file(file_path: Path) -> bool:
        resolved = file_path.resolve()
        if resolved in yielded:
            return False
        if "__pycache__" in file_path.parts:
            return False
        yielded.add(resolved)
        return True

    # root/app/* search
    if app_dir.is_dir():
        for name in common_names:
            file_path = app_dir / name
            if file_path.is_file() and add_file(file_path):
                yield file_path

    # Root-level search
    for name in common_names:
        file_path = root / name
        if file_path.is_file() and add_file(file_path):
            yield file_path

    # Searches common file names anywhere below root
    for name in common_names:
        for file_path in root.rglob(name):
            if add_file(file_path):
                yield file_path

    # Fallback for searching everything else
    for file_path in root.rglob("*.py"):
        if add_file(file_path):
            yield file_path


def find_pyjolt_app_import(
    pyjolt_class: Type,
    root: Optional[Path] = None,
) -> Optional[str]:
    """
    Find the first class that subclasses `pyjolt_class` and return its import string.

    Example return values:
        "app:App"
        "app.main:App"
        "main:App"
    """
    root = (root or Path.cwd()).resolve()
    app_dir = root / "app"

    if not has_python_files(root) and not has_python_files(app_dir):
        raise RuntimeError(
            f"{root} does not appear to contain Python modules in the root or in ./app"
        )

    if str(root) not in sys.path:
        sys.path.insert(0, str(root))

    importlib.invalidate_caches()

    for file_path in iter_candidate_files(root):
        try:
            module_name = path_to_module(file_path, root)
            if not module_name:
                continue
        except ValueError:
            continue

        try:
            module = importlib.import_module(module_name)
        except Exception as exc:
            print(f"Failed to import module '{module_name}': {exc}")
            continue

        for name, obj in inspect.getmembers(module, inspect.isclass):
            if obj.__module__ != module.__name__:
                continue

            try:
                if issubclass(obj, pyjolt_class) and obj is not pyjolt_class:
                    return f"{module_name}:{name}"
            except TypeError:
                continue

    return None


def start(
    cwd: str, debug: bool, app: Optional[str] = None, env_file: Optional[str] = None
):
    """
    Starts application in dev mode.
    """
    from granian import Granian
    from granian.constants import Interfaces, Loops
    from ..patera import Patera

    loaded_env = load_env_file(cwd, env_file)
    if loaded_env is not None:
        print(f"Loaded environment from: {loaded_env}")

    app_path = app
    if app_path is None:
        app_path = find_pyjolt_app_import(Patera, Path(cwd))

    if app_path is None:
        print(
            "Failed to locate Patera implementation. Please specify a correct import string "
            "(example: 'app:App')"
        )
        return
    address: str = os.environ.get("HOST", "127.0.0.1")
    if address == "localhost":
        address = "127.0.0.1"
    _port: int = int(os.environ.get("PORT", 3000))
    _loop = Loops(os.environ.get("GRANIAN_LOOP", "auto"))

    DEFAULT_IGNORE_DIRS = [
        "__dist__",
        "__pycache__",
        "logs",
        "logging",
        ".mypy_cache",
        ".venv",
        ".git",
        ".idea",
        ".vscode",
        "dist",
        "build",
        "node_modules",
    ]
    DECLARED_IGNORE_DIRS = [
        d.strip()
        for d in os.environ.get("GRANIAN_RELOAD_IGNORE_DIRS", "").split(",")
        if d.strip()
    ]
    DEFAULT_IGNORE_DIRS.extend(DECLARED_IGNORE_DIRS)

    DEFAULT_IGNORE_PATTERNS = [
        r".*\.log$",
        r".*\.sqlite$",
        r".*\.sqlite-journal",
        r".*\.db$",
        r".*\.db-journal",
        r".*\.tmp$",
        r".*\.swp$",
    ]
    DECLARED_IGNORE_PATTERNS = [
        p.strip()
        for p in os.environ.get("GRANIAN_RELOAD_IGNORE_PATTERNS", "").split(",")
        if p.strip()
    ]
    DEFAULT_IGNORE_PATTERNS.extend(DECLARED_IGNORE_PATTERNS)

    Granian(
        app_path,
        address=address,
        port=_port,
        interface=Interfaces.ASGI,
        loop=_loop,
        factory=True,
        reload=debug,
        reload_ignore_dirs=DEFAULT_IGNORE_DIRS,
        reload_ignore_patterns=DEFAULT_IGNORE_PATTERNS,
    ).serve()


def start_dev(
    cwd: str, command: str, app: Optional[str] = None, env_file: Optional[str] = None
):
    try:
        start(cwd, True, app, env_file)
    except Exception:
        print("Faled to start Granian dev server. Install patera[dev] if not already.")


def start_prod(
    cwd: str, command: str, app: Optional[str] = None, env_file: Optional[str] = None
):
    start(cwd, False, app, env_file)


def start_cli(
    cwd: str,
    command: str,
    *args,
    app: Optional[str] = None,
    env_file: Optional[str] = None,
    **kwargs,
):

    from ..patera import Patera

    loaded_env = load_env_file(cwd, env_file)
    if loaded_env is not None:
        print(f"Loaded environment from: {loaded_env}")

    app_path = app
    if app_path is None:
        app_path = find_pyjolt_app_import(Patera, Path(cwd))

    if app_path is None:
        print(
            "Failed to locate Patera implementation. Please specify a correct import string "
            "(example: 'app:App')"
        )
        return
    application: Type[Patera] = import_module(app_path)
    app_instance: Patera = application(cli_mode=True)
    command_name = kwargs.pop("command_name", None)
    app_instance.run_cli(command_name, *args, **kwargs)


def start_testing(
    cwd: str,
    command: str,
    *args,
    app: Optional[str] = None,
    env_file: Optional[str] = None,
    **kwargs,
):
    try:
        import pytest  # type: ignore
    except ImportError:
        print(
            "Failed to import Pytest. Please add dependency for running tests. Run: 'pip install patera[testing]'"
        )
        return

    loaded_env = load_env_file(cwd, env_file)
    if loaded_env is not None:
        print(f"Loaded environment from: {loaded_env}")

    pytest_args = kwargs.pop("pytest_args", []) or []

    if not pytest_args or len(pytest_args) == 0:
        pytest_args = [cwd]

    return pytest.main(pytest_args)  # type: ignore



"""
Test client class
"""

from typing import TYPE_CHECKING, cast
from asgi_lifespan import LifespanManager
from httpx import AsyncClient, ASGITransport

if TYPE_CHECKING:
    from ..patera import Patera


class PyJoltTestClient:
    """
    Test client class for testing of Patera applications
    """

    def __init__(self, app: "Patera", base_url: str = "http://testserver"):
        self.app = app
        self._lifespan: LifespanManager = cast(LifespanManager, None)
        self._transport: ASGITransport = cast(ASGITransport, None)
        self.client: AsyncClient = cast(AsyncClient, None)
        self.base_url: str = base_url

    async def __aenter__(self):
        # Starts up app with lifespan events (triggers startup methods)
        self._lifespan = LifespanManager(self.app)
        await self._lifespan.__aenter__()

        self._transport = ASGITransport(app=self.app)
        self.client = AsyncClient(transport=self._transport, base_url=self.base_url)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        # Closes the underlying HTTPX client
        if self.client is not None:
            await self.client.aclose()

        # Executes app shutdown with lifespan events (triggers shutdown methods)
        if self._lifespan is not None:
            await self._lifespan.__aexit__(exc_type, exc_val, exc_tb)

        self.client = cast(AsyncClient, None)
        self._transport = cast(ASGITransport, None)
        self._lifespan = cast(LifespanManager, None)

    async def request(self, method: str, path: str, **kwargs):
        if self.client is None:
            raise RuntimeError(
                "PyJoltTestClient must be used as an async context manager."
            )
        return await self.client.request(method, path, **kwargs)

    async def get(self, path: str, **kwargs):
        return await self.request("GET", path, **kwargs)

    async def post(self, path: str, **kwargs):
        return await self.request("POST", path, **kwargs)

    async def put(self, path: str, **kwargs):
        return await self.request("PUT", path, **kwargs)

    async def patch(self, path: str, **kwargs):
        return await self.request("PATCH", path, **kwargs)

    async def delete(self, path: str, **kwargs):
        return await self.request("DELETE", path, **kwargs)

    async def close(self):
        await self.client.aclose()



"""
Patera application class
"""

# mypy: check-untyped-defs = True
import os
import inspect
import json
from collections.abc import AsyncIterator, Iterable
from enum import StrEnum
from typing import (
    Any,
    Callable,
    Mapping,
    Optional,
    Type,
    TypeVar,
    cast,
    AsyncIterable,
    Union,
)
import aiofiles
from loguru import logger
from werkzeug.exceptions import NotFound, MethodNotAllowed
from pydantic import BaseModel

from jinja2 import (
    Environment,
    FileSystemLoader,
    select_autoescape,
    StrictUndefined,
    Undefined,
)

from patera.media_types import MediaType

from .exceptions.http_exceptions import HtmlAborterException
from .http_statuses import HttpStatus
from .http_methods import HttpMethod
from .request import Request
from .response import Response
from .utilities import get_app_root_path, run_sync_or_async, import_module
from .router import Router
from .static import Static
from .open_api import OpenAPIController
from .controller import path
from .logger import DefaultLogger

from .controller import Controller
from .exceptions import ExceptionHandler
from .base_extension import BaseExtension
from .configuration_base import BaseConfig
from .middleware import MiddlewareBase, AppCallableType
from .cli import CLIController
from .logging.logger_config_base import LoggerBase
from .logging.inmemory_buffer import InMemoryLogBuffer

logger.remove()

PATERA_ASCIIART: str = r"""
  _____        _______ ______ _____
 |  __ \ /\   |__   __|  ____|  __ \     /\
 | |__) /  \     | |  | |__  | |__) |   /  \
 |  ___/ /\ \    | |  |  __| |  _  /   / /\ \
 | |  / ____ \   | |  | |____| | \ \  / ____ \
 |_| /_/    \_\  |_|  |______|_|  \_\/_/    \_\
A Fast, Simple, and Productive Python Web Framework
"""

PATERA_VERSION: str = "0.111.x"

T = TypeVar("T", bound="Patera")


def app_path(url_path: Optional[str] = None) -> Callable[[Type[T]], Type[T]]:
    def decorator(cls: Type[T]) -> Type[T]:
        setattr(cls, "_base_url_path", url_path)
        return cls

    return decorator


def app(import_name: str, configs: Type[BaseConfig]) -> Callable[[Type[T]], Type[T]]:
    def decorator(cls: Type[T]) -> Type[T]:
        setattr(cls, "_app_configs", {"import_name": import_name, "configs": configs})
        return cls

    return decorator


def on_startup(func) -> Callable:
    """
    Decorated methods will run in alphabetical order on app startup
    """
    setattr(func, "_on_startup_method", True)
    return func


def on_shutdown(func) -> Callable:
    """
    Decorated methods will run in alphabetical order on app shutdown
    """
    setattr(func, "_on_shutdown_method", True)
    return func


class ScopeType(StrEnum):
    LIFESPAN = "lifespan"
    HTTP = "http"
    WEBSOCKET = "websocket"


class MissingAppConfigurations(Exception):
    def __init__(
        self,
        msg: str = (
            "Missing application configurations. "
            "Please make sure to use the @app_configs "
            "decorator with appropriate arguments."
        ),
    ):
        super().__init__(msg)


class MissingImportModule(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class WrongModuleLoadType(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


def validate_config(config_obj_or_type: Type[BaseConfig] | BaseConfig) -> BaseConfig:
    # If it's already an instance
    if isinstance(config_obj_or_type, BaseConfig):
        return cast(BaseConfig, config_obj_or_type)  # already validated by Pydantic

    if inspect.isclass(config_obj_or_type) and issubclass(
        config_obj_or_type, BaseConfig
    ):
        try:
            instance = config_obj_or_type()  # type: ignore[call-arg]
        except Exception as e:
            raise MissingAppConfigurations(
                f"Could not instantiate config class {config_obj_or_type.__name__}: {e}"
            ) from e
        return BaseConfig.model_validate(instance.model_dump())

    raise MissingAppConfigurations("Configs must be a subclass of patera.BaseConfig.")


def inherits_from(class_obj_or_instance, base_name: str) -> bool:
    # Checks inheritance via string comparison
    if inspect.isclass(class_obj_or_instance):
        return any(base.__name__ == base_name for base in class_obj_or_instance.__mro__)
    return any(
        base.__name__ == base_name for base in class_obj_or_instance.__class__.__mro__
    )


class Patera:
    """Patera class implementation. Used to create a new application instance"""

    def __init__(self, cli_mode: bool = False):
        """Init function"""
        app_configs: dict[str, str | object | dict] | None = getattr(
            self.__class__, "_app_configs", None
        )
        if app_configs is None:
            raise MissingAppConfigurations()

        import_name: str = cast(str, app_configs.get("import_name", None))
        configs: Type[BaseConfig] = cast(
            Type[BaseConfig], app_configs.get("configs", None)
        )
        if configs is None or not issubclass(configs, BaseConfig):
            raise MissingAppConfigurations(
                "Missing valid configs object in @app_configs. "
                "Configuration class must inherit from patera.BaseConfig"
            )
        self._this_path = os.path.dirname(__file__)
        self._app_base_url: str = getattr(self.__class__, "_base_url_path", "")
        self._is_built = False
        self._root_path = get_app_root_path(import_name)
        # Dictionary which holds application configurations
        validated_configs: BaseConfig = validate_config(configs)
        self._configs = {**validated_configs.model_dump()}
        static_dir = self.get_conf("STATIC_DIR").lstrip("/\\")
        self._static_files_path = os.path.join(self._root_path, static_dir)
        self._templates_path = self._root_path + self.get_conf("TEMPLATES_DIR")

        self._all_templates_paths = [self._templates_path]

        self._url_for_alias: dict[str, str] = {
            self.get_conf("STATIC_CONTROLLER_NAME"): "Static.get"
        }
        self._logger_sink_ids: list[int] = []

        # creates Jinja2 environment for entire app
        self._jinja_environment = Environment(
            loader=None,
            autoescape=select_autoescape(["html", "xml"]),
            undefined=StrictUndefined
            if self.get_conf("TEMPLATES_STRICT", True)
            else Undefined,
            auto_reload=self.get_conf("AUTO_RELOAD", False),
            enable_async=True,
        )
        sink_id = DefaultLogger(self).configure()
        self._logger_sink_ids.append(sink_id)

        self._router = Router(self.get_conf("STRICT_SLASHES", False))
        self._socket_router = Router(self.get_conf("STRICT_SLASHES", False))
        self._logger = logger

        self.log_buffer = InMemoryLogBuffer(
            maxlen=self._configs.get("IN_MEMORY_LOG_BUFFER_SIZE", 1)
        )
        # Capture everything (TRACE and above) in the in-memory log buffer
        self._log_buffer_sink_id = logger.add(
            self.log_buffer, level="TRACE", enqueue=True, backtrace=True, diagnose=False
        )
        self._logger_sink_ids.append(self._log_buffer_sink_id)

        self._app: AppCallableType = self._base_app
        self._middleware: list[Callable] = []
        self._controllers: dict[str, "Controller"] = {}
        self._cli_controllers: dict[str, "CLIController"] = {}
        self._exception_handlers: dict[str, Callable] = {}
        self._json_spec: Optional[dict] = None
        self._db_models: dict[str, list[Type]] = {}
        self._db_name_configs_map: dict[str, str] = {}

        self._extensions: dict = {}
        self.global_context_methods: list[Callable] = []

        self._on_startup_methods: list[Callable] = []
        self._on_shutdown_methods: list[Callable] = []

        self._get_startup_methods()
        self._get_shutdown_methods()

        models: Optional[list[str]] = self.get_conf("MODELS", None)
        controllers: Optional[list[str]] = self.get_conf("CONTROLLERS", None)
        cli_controllers: Optional[list[str]] = self.get_conf("CLI_CONTROLLERS", None)
        exception_handlers: Optional[list[str]] = self.get_conf(
            "EXCEPTION_HANDLERS", None
        )
        extensions: Optional[list[str]] = self.get_conf("EXTENSIONS", None)
        middleware: Optional[list[str]] = self.get_conf("MIDDLEWARE", None)
        loggers: Optional[list[str]] = self.get_conf("LOGGERS", None)
        # if NOT in CLI mode (cli_mode = False)
        # all extensions, models, controllers, exception handlers and middleware
        # is registered and configured with the app.
        if not cli_mode:
            self._enable_cors()  # enables CORS middleware if configured
            self._load_modules(loggers)
            self._load_modules(models)
            self._load_modules(extensions)
            self._load_modules(controllers)
            self._load_modules(cli_controllers)
            self._load_modules(exception_handlers)
            self._load_modules(middleware)
        # if in CLI mode only models and extension are registered
        # and configured with the app
        else:
            self._load_modules(models)
            self._load_modules(extensions)
            self._load_modules(cli_controllers)

        self._jinja_environment.loader = FileSystemLoader(self._all_templates_paths)

    def _enable_cors(self):
        cors_enabled: bool = self.get_conf("CORS_ENABLED", True)
        if not cors_enabled:
            return

        # pylint: disable-next=C0415
        from .cors.cors_mw import CORSMiddleware

        self.logger.info(f"Registering middleware: {CORSMiddleware.__name__}")
        self._middleware.append(
            # pylint: disable-next=W0108
            lambda app, next_app: CORSMiddleware(app, next_app)
        )

    def _load_modules(self, modules: Optional[list[str]] = None):
        if modules is None:
            return
        for import_string in modules:
            obj = import_module(import_string)
            if obj is None:
                raise MissingImportModule(
                    f"Failed to load module: {import_string}. Check path in configurations."
                )
            if inspect.isclass(obj) and issubclass(obj, Controller):
                self.logger.info(f"Registering controller: {obj.__name__}")
                self.register_controller(obj)
                continue
            if inspect.isclass(obj) and issubclass(obj, ExceptionHandler):
                self.logger.info(f"Registering exception handler: {obj.__name__}")
                self.register_exception_handler(obj)
                continue
            if inherits_from(obj, "SqlDatabase"):
                self.logger.info(
                    f"Initilizing database: {obj.__class__.__name__} ({obj.configs_name})"
                )
                obj.init_app(self)
                self._db_name_configs_map[obj.db_name] = obj.configs_name
                continue
            if inherits_from(obj, "BaseExtension"):
                ext_name = ""
                conf_name = ""
                if isinstance(obj, BaseExtension):
                    obj.init_app(self)
                    ext_name = obj.__class__.__name__
                    conf_name = obj.configs_name
                else:
                    # if passed extension is not yet initilized
                    obj_inst = obj()
                    obj_inst.init_app(self)
                    ext_name = obj_inst.__class__.__name__
                    conf_name = obj_inst.configs_name
                self.logger.info(f"Initilizing extension: {ext_name} ({conf_name})")
                continue
            if inspect.isclass(obj) and inherits_from(obj, "DeclarativeBaseModel"):
                self.logger.info(f"Loaded database model: {obj.__name__}")
                if obj.db_name() not in self._db_models:
                    self._db_models[obj.db_name()] = []
                self._db_models[obj.db_name()].append(obj)
                continue
            if inherits_from(obj, "CLIController"):
                cli_controller = obj(self)
                self.register_cli_controller(cli_controller)
                self.logger.info(f"Registering cli controller: {obj.__name__}")
                continue
            if inspect.isclass(obj) and issubclass(obj, MiddlewareBase):
                self.logger.info(f"Registering middleware: {obj.__name__}")
                self._middleware.append(
                    lambda app, next_app, mdlwr_class=obj: mdlwr_class(app, next_app)
                )
                continue
            if inspect.isclass(obj) and issubclass(obj, LoggerBase):
                sink_id = obj(self).configure()
                self._logger_sink_ids.append(sink_id)
                print(f"Registering logger: {obj.__name__}")
                continue
            raise WrongModuleLoadType(
                f"Failed to load module {obj.__name__ or obj.__class__.__name__}. Extensions can be passed as instances or classes (if custom config names are not used and instance is not used elsewhere). Controllers, cli controllers, exception handlers and middleware as classes."
            )

    def _get_startup_methods(self):
        methods = []
        for name in dir(self):
            method = getattr(self, name)
            if callable(method) and getattr(method, "_on_startup_method", False):
                methods.append((name, method))
        methods.sort(key=lambda x: x[0])  # by method name
        self._on_startup_methods = [m for _, m in methods]

    def _get_shutdown_methods(self):
        methods = []
        for name in dir(self):
            method = getattr(self, name)
            if callable(method) and getattr(method, "_on_shutdown_method", False):
                methods.append((name, method))
        methods.sort(key=lambda x: x[0])  # by method name
        self._on_shutdown_methods = [m for _, m in methods]

    def get_conf(self, config_name: str, default: Any = None) -> Any:
        """
        Returns app configuration with provided config_name.
        Raises error if configuration is not found.
        """
        return self.configs.get(config_name, default)

    def add_global_context_method(self, func: Callable):
        """
        Adds global context method to global_context_methods array
        """
        self.global_context_methods.append(func)

    async def _base_app(self, req: Request) -> Response:
        """
        The bare-bones application without any middleware.
        Calls the route handler directly.
        """
        res: Response = await run_sync_or_async(
            req.route_handler, req, **req.route_parameters
        )
        return res

    async def abort_route_not_found(
        self, send, req: Request, path_data: Mapping[str, Any]
    ):
        """
        Aborts request because route was not found
        """
        exc: NotFound | MethodNotAllowed | None = path_data.get("exc")
        if exc is not None:
            handler: Callable | None = (
                self._exception_handlers.get(exc.__class__.__name__, None) or None
            )
            if handler:
                res = await run_sync_or_async(handler, req, exc)
                response_type = res.expected_body_type() or exc.__class__
                return await self.send_response(res, send, response_type)
        ##sends generic response if custom handler not available
        await send(
            {
                "type": "http.response.start",
                "status": 404,
                "headers": [(b"content-type", b"application/json")],
            }
        )
        await send(
            {
                "type": "http.response.body",
                "body": b'{ "status": "error", "message": "Endpoint not found" }',
            }
        )

    async def _iterate_stream(
        self,
        iterable: Union[AsyncIterable[bytes], Iterable[bytes]],
    ) -> AsyncIterator[bytes]:
        """
        Normalizes async/sync iterables into an async iterator of bytes.
        Accepts bytes/bytearray/str chunks and encodes/normalizes to bytes.
        """

        async def _aiter_from_sync(sync_iter: Iterable[bytes]) -> AsyncIterator[bytes]:
            for chunk in sync_iter:
                yield self._normalize_chunk(chunk)

        if hasattr(iterable, "__aiter__"):
            async for chunk in iterable:  # type: ignore[attr-defined]
                yield self._normalize_chunk(chunk)
        else:
            async for chunk in _aiter_from_sync(iterable):  # type: ignore[arg-type]
                yield chunk

    def _normalize_chunk(self, chunk: Any) -> bytes:
        if isinstance(chunk, (bytes, bytearray)):
            return bytes(chunk)
        if isinstance(chunk, str):
            return chunk.encode("utf-8")
        raise TypeError(
            f"Streaming chunks must be bytes, bytearray or str, got {type(chunk)!r}"
        )

    async def send_response(
        self, res: Response, send, response_type: Optional[Type[Any]] = None
    ):
        """
        Sends response
        """
        # Build headers for ASGI send
        headers = []
        for k, v in res.headers.items():
            headers.append((k.encode("utf-8"), v.encode("utf-8")))
        await send(
            {
                "type": "http.response.start",
                "status": res.status_code.value
                if isinstance(res.status_code, HttpStatus)
                else res.status_code,
                "headers": headers,
            }
        )
        # Zero-copy _parameters_ were stashed in res._zero_copy
        if res.zero_copy is not None:
            params = res.zero_copy
            file_path = params["file_path"]
            start = params["start"]
            length = params["length"]

            # stream in 1 MiB chunks
            chunk_size = 1 * 1024 * 1024
            remaining = length

            async with aiofiles.open(file_path, "rb") as f:
                await f.seek(start)
                while remaining > 0:
                    to_read = min(remaining, chunk_size)
                    chunk = await f.read(to_read)
                    if not chunk:
                        break
                    remaining -= len(chunk)
                    await send(
                        {
                            "type": "http.response.body",
                            "body": chunk,
                            "more_body": remaining > 0,
                        }
                    )
            return

        if getattr(res, "is_streaming", False) and res.is_streaming:
            stream_iter = res.stream_iterable
            if stream_iter is None:
                await send(
                    {"type": "http.response.body", "body": b"", "more_body": False}
                )
                return

            async for chunk in self._iterate_stream(stream_iter):
                await send(
                    {
                        "type": "http.response.body",
                        "body": chunk,
                        "more_body": True,
                    }
                )

            # Final empty chunk with more_body=False
            await send(
                {
                    "type": "http.response.body",
                    "body": b"",
                    "more_body": False,
                }
            )
            return

        if res.body and res.content_type in [
            MediaType.APPLICATION_JSON,
            MediaType.APPLICATION_PROBLEM_JSON,
            MediaType.APPLICATION_X_NDJSON,
        ]:
            if (
                response_type
                and issubclass(response_type, BaseModel)
                and isinstance(res.body, dict)
            ):
                res.body = response_type(**res.body).model_dump_json().encode("utf-8")
            elif (
                response_type
                and issubclass(response_type, BaseModel)
                and isinstance(res.body, BaseModel)
            ):
                res.body = res.body.model_dump_json().encode("utf-8")
            elif res.body and response_type is None and isinstance(res.body, BaseModel):
                res.body = res.body.model_dump_json().encode("utf-8")
            elif not response_type and isinstance(res.body, dict):
                # tries to serialize plain dict to json. Works if the dict contains only json serializable types
                res.body = json.dumps(res.body).encode("utf-8")
            elif res.body and not isinstance(res.body, (bytes, bytearray)):
                # tries to serialize other types (not bytes or bytesarray) to json
                res.body = json.dumps(res.body).encode("utf-8")

        await send(
            {
                "type": "http.response.body",
                "body": res.body or b"",
            }
        )

    async def _lifespan_app(self, _, receive, send):
        """This loop will listen for 'startup' and 'shutdown'"""
        while True:
            message = await receive()

            if message["type"] == "lifespan.startup":
                for method in self._on_startup_methods:
                    await run_sync_or_async(method)
                await send({"type": "lifespan.startup.complete"})

            elif message["type"] == "lifespan.shutdown":
                for method in self._on_shutdown_methods:
                    await run_sync_or_async(method)
                for logger_sink_id in self._logger_sink_ids:
                    self.logger.remove(logger_sink_id)
                await send({"type": "lifespan.shutdown.complete"})
                return  # Exit the lifespan loop

    async def _handle_http_request(self, scope, receive, send):
        """
        Handles http requests
        """
        method: str = scope["method"].upper()
        url_path: str = scope["path"]
        self._log_request(scope, method, url_path)

        route_handler, path_kwargs = self.router.match(url_path, method)
        req = self.request_class(
            scope, receive, self, path_kwargs, cast(Callable, route_handler)
        )

        if not route_handler:
            return await self.abort_route_not_found(send, req, path_kwargs)

        try:
            try:
                res: Response = await self._app(req)
                if not isinstance(res, Response):
                    # pylint: disable-next=W0719
                    raise Exception(
                        "Return object of request handlers must be an instance of Response"
                    )
                response_type: Optional[Type[Any]] = req.response.expected_body_type()
                return await self.send_response(res, send, response_type)
            except HtmlAborterException as exc:
                res = (await req.res.html(exc.template, context=exc.data)).status(
                    exc.status_code
                )
                return await self.send_response(res, send, None)
            # pylint: disable-next=W0718
            except Exception as exc:
                handler = (
                    self._exception_handlers.get(exc.__class__.__name__, None) or None
                )
                if not handler:
                    # pylint: disable-next=W0719
                    raise
                res = await run_sync_or_async(handler, req, exc)
                response_type = res.expected_body_type() or exc.__class__
                return await self.send_response(res, send, response_type)
        # pylint: disable-next=W0718
        except Exception as exc:
            # Catches every error and returns internal server error message
            # if the app is in production (DEBUG = False)
            # else reraises the error
            if not self.get_conf("DEBUG", False):
                res = req.res.json(
                    {
                        "status": "error",
                        "message": "Internal server error",
                    }
                ).status(HttpStatus.INTERNAL_SERVER_ERROR)
                self.logger.critical(
                    f"Unhandled critical error: ({req.method}) {req.path}, {req.route_parameters}"
                )
                return await self.send_response(res, send, exc.__class__)
            raise

    def _log_request(self, scope, method: str, url_path: str) -> None:
        """
        Logs incoming request
        """
        logger.info(
            f"HTTP request. CLIENT: {(scope.get('client') or ('-', ''))[0]}, SCHEME: {scope['scheme']}, METHOD: {method}, PATH: {url_path}, QUERY_STRING: {scope['query_string'].decode('utf-8')}"
        )

    def register_static_controller(self, base_path: str):
        print("Register static: ", base_path)
        static_controller_dec = path(f"{base_path}", open_api_spec=False)
        static_controller = static_controller_dec(Static)
        self.register_controller(static_controller, with_base_path=False)  # type: ignore

    def register_openapi_controller(self):
        openapi_controller_dec = path(
            self.get_conf("OPEN_API_URL"), open_api_spec=False
        )
        openapi_controller = openapi_controller_dec(OpenAPIController)
        self.register_controller(openapi_controller)

    def build(self) -> None:
        """
        Build the final app by wrapping self._app in all middleware.
        Apply them in reverse order so the first middleware in the list
        is the outermost layer.
        """
        print(PATERA_ASCIIART)
        print(f"Starting Patera {PATERA_VERSION} application '{self.app_name}'")
        self.register_static_controller(self.get_conf("STATIC_URL"))
        if self.get_conf("OPEN_API", False):
            self.build_openapi_spec()
            self.register_openapi_controller()
        built_app: AppCallableType = self._base_app
        for factory in reversed(self._middleware):
            built_app = factory(self, built_app)
        self._app = built_app
        self._is_built = True

    def add_extension(self, extension):
        """
        Adds extension to extension map
        """
        ext_name = extension.configs_name
        self._extensions[ext_name] = extension

    def activate_extension(self, extension: "Type[BaseExtension]"):
        extension_instance = extension()
        extension_instance.init_app(self)

    def _add_route_function(
        self, method: str, url_path: str, func: Callable, endpoint_name: str
    ):
        """
        Adds the function to the Router.
        Raises DuplicateRoutePath if a route with the same (method, path) is already registered.
        """
        try:
            if method == HttpMethod.SOCKET.value:
                self._socket_router.add_route(url_path, func, [method], endpoint_name)
            else:
                self.router.add_route(url_path, func, [method], endpoint_name)
        except Exception as e:
            raise e

    def register_controller(
        self, *ctrls: "type[Controller]", with_base_path: bool = True
    ):
        """Registers controller class with application"""
        base_path: str = self._app_base_url if with_base_path else ""
        for ctrl in ctrls:
            dev_only: bool = getattr(ctrl, "_development", False)
            if dev_only and not self.get_conf("DEBUG", False):
                continue
            ctrl_path: str = getattr(ctrl, "_controller_path")
            ctrl_open_api_spec = getattr(ctrl, "_include_open_api_spec")
            ctrl_open_api_tags = getattr(ctrl, "_open_api_tags", None)
            ctrl_instance = ctrl(
                self, ctrl_path, ctrl_open_api_spec, ctrl_open_api_tags
            )

            self._controllers[ctrl_instance.path] = ctrl_instance
            endpoint_methods: dict[str, dict[str, str | Callable]] = (
                ctrl_instance.get_endpoint_methods()
            )
            for http_method, endpoints in endpoint_methods.items():
                for url_path, method in endpoints.items():
                    method_name: Callable = method["method"].__name__  # type: ignore
                    # handler: Callable = getattr(ctrl_instance, method_name) # type: ignore
                    endpoint_name: str = (
                        f"{ctrl_instance.__class__.__name__}.{method_name}"
                    )
                    self._add_route_function(
                        http_method,
                        base_path + ctrl_instance.path + url_path,
                        cast(Callable, cast(dict, method)["method"]),
                        endpoint_name,
                    )

    def register_cli_controller(self, ctrl: CLIController) -> None:
        self._cli_controllers[ctrl.ctrl_name] = ctrl

    def register_exception_handler(self, *handlers: "type[ExceptionHandler]"):
        """Registers exception controller with application"""
        for handler in handlers:
            handler_instance = handler(self)
            handled_exceptions = handler_instance.get_exception_mapping()
            self._exception_handlers.update(handled_exceptions)

    def url_for(self, endpoint: str, **values) -> str:
        """
        Returns url for endpoint method
        :param endpoint: the name of the endpoint handler method namespaced
        with the controller name
        :param values: dynamic route parameters
        :return: url (string) for endpoint
        """
        endpoint = self._url_for_alias.get(endpoint, endpoint)
        adapter = self.router.url_map.bind("")  # Binds map to base url
        # If a value starts with a forward slash, systems like MacOS/Linux treat it as an absolute path
        try:
            return adapter.build(endpoint, values)
        except NotFound as exc:
            raise ValueError(f"Endpoint '{endpoint}' does not exist.") from exc
        except MethodNotAllowed as exc:
            raise ValueError(
                f"Endpoint '{endpoint}' exists but does not allow the method."
            ) from exc
        except Exception as exc:
            raise ValueError(
                f"Error building URL for endpoint '{endpoint}': {exc}"
            ) from exc

    def build_openapi_spec(self):
        """Builds open api spec"""
        # pylint: disable-next=C0415
        from .open_api import build_openapi

        self._json_spec = build_openapi(
            self._controllers,
            title=self.app_name,
            version=self.version,
            openapi_version="3.0.3",
            servers=[f"http://localhost:{self.get_conf('PORT', 3000)}"],
        )

    def add_on_startup_method(self, func: Callable):
        """
        Adds method to on_startup collection
        """
        self._on_startup_methods.append(func)

    def add_on_shutdown_method(self, func: Callable):
        """
        Adds method to on_shutdown collection
        """
        self._on_shutdown_methods.append(func)

    def register_alias(self, alias: str, endpoint: str):
        """
        Registers an alias for an endpoint name.
        Useful for url_for lookups.
        """
        self._url_for_alias[alias] = endpoint

    def run_cli(self, command_name: str, *args, **kwargs) -> None:
        """
        Executes the registered CLI commands.
        """
        if command_name is None or ":" not in command_name:
            print(
                "Invalid command name. Command name must be of the format CLIController:Command"
            )
            return
        ctrl_name, command = command_name.split(":", 1)
        ctrl = cast(CLIController, self._cli_controllers.get(ctrl_name, None))
        if ctrl is None:
            print(f"CLI controller with name {ctrl_name} was not found")

        method = ctrl.find_method(command)
        if method is None:
            print(
                f"CLI method with name {command} in controller {ctrl_name} not found."
            )
            return
        ctrl.run_command(method, *args, **kwargs)
        return

    def add_template_path(self, path: str):
        """Adds a template path"""
        self._all_templates_paths.append(path)

    @property
    def json_spec(self) -> dict | None:
        return self._json_spec

    @property
    def router(self) -> Router:
        """Router instance property of the app"""
        return self._router

    @property
    def configs(self) -> dict[str, Any]:
        """
        Returns configuration dictionary
        """
        return self._configs

    @property
    def root_path(self) -> str:
        """
        Returns root path of application
        """
        return self._root_path

    @property
    def app(self):
        """
        Returns self
        For compatibility with the Controller class
        which contains the app object on the app property
        """
        return self

    @property
    def static_files_path(self) -> str:
        """Static files paths"""
        return self._static_files_path

    @property
    def version(self) -> str:
        return self.get_conf("VERSION")

    @property
    def app_name(self) -> str:
        return self.get_conf("APP_NAME")

    @property
    def logger(self):
        return self._logger

    @property
    def jinja_environment(self) -> Environment:
        return self._jinja_environment

    @property
    def request_class(self) -> Type[Request]:
        """
        Returns the Request class used by the application.
        Can be overridden in configs to provide a custom Request subclass.
        """
        return self.get_conf("REQUEST_CLASS", Request)

    @property
    def response_class(self) -> Type[Response]:
        """
        Returns the Response class used by the application.
        Can be overridden in configs to provide a custom Response subclass.
        """
        return self.get_conf("RESPONSE_CLASS", Response)

    @property
    def extensions(self) -> "dict[str, BaseExtension]":
        """Returns dictionary with all registered extensions"""
        return self._extensions

    async def __call__(self, scope, receive, send):
        """
        Once built, __call__ just delegates to the fully wrapped app.
        """
        if not self._is_built:
            self.build()
        if scope["type"] == ScopeType.LIFESPAN.value:
            return await self._lifespan_app(scope, receive, send)
        if scope["type"] == ScopeType.HTTP.value:
            return await self._handle_http_request(scope, receive, send)
        if scope["type"] == ScopeType.WEBSOCKET.value:
            return await self._handle_websocket_request(scope, receive, send)
        raise ValueError(f"Unsupported scope type {scope['type']}")

    async def _handle_websocket_request(self, scope, receive, send):
        """
        Handles websocket requests
        """
        method: str = "SOCKET"
        url_path: str = scope["path"]
        self._log_request(scope, method, url_path)

        route_handler, path_kwargs = self._socket_router.match(url_path, method)
        if not route_handler:
            await send({"type": "websocket.close", "code": 1000})
            return
        req = Request(scope, receive, self, path_kwargs, cast(Callable, route_handler))
        req.set_send(send)
        try:
            await run_sync_or_async(route_handler, req, **path_kwargs)
        # pylint: disable-next=W0718
        except Exception as exc:
            await send(
                {
                    "type": "websocket.close",
                    "code": 1011,
                    "reason": "Internal server error",
                }
            )
            self.logger.critical(
                f"Unhandled critical error in websocket: ({req.method}) {req.path}, {req.route_parameters}: {exc}"
            )
            raise exc
